/*
   esp32-cam-pan-tilt
   procedury sterowania serwomechanizmami ruchomej głowicy
   napisał rysiek ars.info.pl
*/

/*
   dołączona do cmd_handler obsługa przesyłanych komend GET
   związanych ze sterowaniem serwami
   przesłanie: /control?var=<key>&val=<val>  /control?var=s_left&val=0
   składnia:
   variable s_left, s_right, s_up, s_down
   value 0 (w tej wersji bez znaczenia)
*/
#include <string.h>
#include <Servo.h>
#include <soc/soc.h>
#include "soc/rtc_cntl_reg.h"

#include "FS.h"
#include "SPIFFS.h"
#include "src/jsonlib/jsonlib.h"

#include "uTimerLib.h"


#define PREFERENCES_SERVO_MAX_SIZE 500
#define PREFERENCES_SERVO_FILE "/serwo-pref.json"
void dumpPrefsServo(fs::FS &fs);
void removePrefsServo(fs::FS &fs);
void loadPrefsServo(fs::FS &fs);
void savePrefsServo(fs::FS &fs);
extern bool filesystem;

//%%obsługa serw%%
Servo servo1;  // create servo object to control a servo
Servo servo2;  // create servo object to control a servo
// twelve servo objects can be created on most boards

// GPIO the servo is attached to
static const int servo1Pin = 15;//14;
static const int servo2Pin = 14;//15;
static const int servaOnOff= 13;
static const int servaOnOff1= 12;


int servo1Pos = 90;
int servo2Pos = 90;
int s_horizon_min=0,s_horizon_max=180,s_horizon_cen=90,s_vertic_min=0,s_vertic_max=180,s_vertic_cen=90;

bool serva_miekki_ruch_cnl=false, servo1_ruch=false, servo2_ruch=false;
unsigned long serva_mr_interwal;
void serwaMiekkiRuch(void);
bool serwaMR_krok_ruchu(int *p_servoPos);
void servaMiekkiRuch_start(int *p_servoPos);

/*
 * sterowanie ruchami serwo
 */
bool komendyGET_serwomechanizmow(char *variable, char *value_t)
{
int value;

  value =atoi(value_t);

  if (!strcmp(variable, "s_stop"))
  {
    Serial.println("STOP");

    return 0;
  }
  
  if (!strcmp(variable, "s_horizon"))
  {
    if (value >= 0 && value <=180) {
      servo2Pos =value;
      servaMiekkiRuch_start(&servo2Pos);  //servo2.write(servo2Pos);
    //  servo2.write(servo2Pos);
    }
    Serial.println(servo2Pos);
    Serial.println("Horizontal pos.");

    return 0;
  } else if (!strcmp(variable, "s_vertic"))
  {
    if (value >= 0 && value <=180) {
      servo1Pos =value;
      servaMiekkiRuch_start(&servo1Pos);  //servo1.write(servo1Pos);
      //servo1.write(servo1Pos);
    }
    Serial.println(servo1Pos);
    Serial.println("Vertical pos.");

    return 0;
  } else return 1;
}

/*
 * komendy ustawień ograniczeń ruchów serwo
 */
bool komendy_ograniczen_serwo(char *variable, char *value_t)
{
int value;
bool error_flag=1;

  value =atoi(value_t);

  while (value >= 0 && value <=180)
  {
    if (!strcmp(variable, "s_horizon_min"))
    {
      if (value<=s_horizon_max){
        error_flag=0;
        s_horizon_min =value;
        break;
      }
    }
    if (!strcmp(variable, "s_horizon_max"))
    {
      if (value>=s_horizon_min){
        error_flag=0;
        s_horizon_max=value;
        break;        
      }
    }
    if (!strcmp(variable, "s_horizon_cen"))
    {
      if (value>=s_horizon_min && value<=s_horizon_max){
        error_flag=0;
        s_horizon_cen=value;
        break;                
      }
    }
    if (!strcmp(variable, "s_vertic_min"))
    {
      if (value<=s_vertic_max){
        error_flag=0;
        s_vertic_min =value;
        break;
      }
    }
    if (!strcmp(variable, "s_vertic_max"))
    {
      if (value>=s_vertic_min){
        error_flag=0;
        s_vertic_max =value;
        break;        
      }
    }
    if (!strcmp(variable, "s_vertic_cen"))
    {
      if (value>=s_vertic_min && value<=s_vertic_max){
        error_flag=0;
        s_vertic_cen =value;
        break;                
      }
    }
   // if (variable==NULL || value==NULL){
      Serial.print("s_horizon_min "); Serial.println(s_horizon_min);
      Serial.print("s_horizon_max "); Serial.println(s_horizon_max);
      Serial.print("s_horizon_cen "); Serial.println(s_horizon_cen);
      Serial.print("s_vertic_min "); Serial.println(s_vertic_min);
      Serial.print("s_vertic_max "); Serial.println(s_vertic_max);
      Serial.print("s_vertic_cen "); Serial.println(s_vertic_cen);
      
      return 0;
  //  }
  }
  if (error_flag!=0){
    Serial.println("ERROR Vertical or Horizontal restriction");
    return 1;
  }
  Serial.println(variable);
  Serial.println(value);
  return 0;
}
 
/*
   inicjacja serwomechanizmów
*/
void setup_servo(void)
{
  WRITE_PERI_REG(RTC_CNTL_BROWN_OUT_REG, 0); //disable brownout detector
  //  servo1.setPeriodHertz(50);    // standard 50 hz servo
  //  servo2.setPeriodHertz(50);    // standard 50 hz servo
  //  servoN1.attach(2, 1000, 2000);
  //  servoN2.attach(13, 1000, 2000);

  //  servo1.attach(SERVO_1, 1000, 2000);
  //  servo2.attach(SERVO_2, 1000, 2000);
  if (filesystem) {
      loadPrefsServo(SPIFFS);
  } else {
      Serial.println("No Internal Filesystem, cannot load or save preferences");
  }

//włączanie zasilania serw
  pinMode(servaOnOff, OUTPUT);
  digitalWrite(servaOnOff, HIGH);
  pinMode(servaOnOff1, OUTPUT);  
  digitalWrite(servaOnOff1, HIGH);
  


  servo1Pos=s_vertic_cen;
  servo2Pos=s_horizon_cen;
  servo1.attach(servo1Pin);  // attaches the servo on the servoPin to the servo object
  servo1.write(servo1Pos);
  delay(700);
  servo2.attach(servo2Pin);  // attaches the servo on the servoPin to the servo object
  servo2.write(servo2Pos);
  delay(700);  
}

void odczyt_zmiennych_zewnetrznych(char *d)
{
  d+= sprintf(d,"function odczyt_zmiennych_zewnetrznych(){\n");
  d+= sprintf(d,"x_min=%d;\nx_max=%d;\nx_cen=%d;\n",s_vertic_min,s_vertic_max,s_vertic_cen);
  d+= sprintf(d,"y_min=%d;\ny_max=%d;\ny_cen=%d;\n",s_horizon_min,s_horizon_max,s_horizon_cen);
  d+= sprintf(d,"x=%d;\ny=%d;\n",servo1Pos,servo2Pos);
  d+= sprintf(d,"}\n");  
  *d++ =0;  
}

//procedury SPIFFS dla serw

void dumpPrefsServo(fs::FS &fs){
  if (fs.exists(PREFERENCES_SERVO_FILE)) {
    // Dump contents for debug
    File file = fs.open(PREFERENCES_SERVO_FILE, FILE_READ);
    int countSize = 0;
    while (file.available() && countSize <= PREFERENCES_SERVO_MAX_SIZE) {
        Serial.print(char(file.read()));
        countSize++;
    }
    Serial.println("");
    file.close();
  } else {
    Serial.printf("%s not found, nothing to dump.\r\n", PREFERENCES_SERVO_FILE);
  }
}

void removePrefsServo(fs::FS &fs) {
  if (fs.exists(PREFERENCES_SERVO_FILE)) {
    Serial.printf("Removing %s\r\n", PREFERENCES_SERVO_FILE);
    if (!fs.remove(PREFERENCES_SERVO_FILE)) {
      Serial.println("Error removing preferences");
    }
  } else {
    Serial.println("No saved servo preferences file to remove");
  }
}

void loadPrefsServo(fs::FS &fs){
  if (fs.exists(PREFERENCES_SERVO_FILE)) {
    // read file into a string
    String prefs;
    Serial.printf("Loading preferences from file %s\r\n", PREFERENCES_SERVO_FILE);
    File file = fs.open(PREFERENCES_SERVO_FILE, FILE_READ);
    if (!file) {
      Serial.println("Failed to open servo preferences file for reading, maybe corrupt, removing");
      removePrefsServo(SPIFFS);
      return;
    }
    size_t size = file.size();
    if (size > PREFERENCES_SERVO_MAX_SIZE) {
      Serial.println("Preferences servo file size is too large, maybe corrupt, removing");
      removePrefsServo(SPIFFS);
      return;
    }
    while (file.available()) {
        prefs += char(file.read());
        if (prefs.length() > size) {
          // corrupted SPIFFS files can return data beyond their declared size.
          Serial.println("Preferences servo file failed to load properly, appears to be corrupt, removing");
          removePrefsServo(SPIFFS);
          return;
        }
    }
    // get servo reference
    s_horizon_min =jsonExtract(prefs, "s_horizon_min").toInt();
    s_horizon_max =jsonExtract(prefs, "s_horizon_max").toInt();
    s_horizon_cen =jsonExtract(prefs, "s_horizon_cen").toInt();
    s_vertic_min =jsonExtract(prefs, "s_vertic_min").toInt();
    s_vertic_max =jsonExtract(prefs, "s_vertic_max").toInt();
    s_vertic_cen =jsonExtract(prefs, "s_vertic_cen").toInt();
    
    // close the file
    file.close();
    dumpPrefsServo(SPIFFS);
  } else {
    Serial.printf("Preference file %s not found; using system defaults.\r\n", PREFERENCES_SERVO_FILE);
  }
}

void savePrefsServo(fs::FS &fs){
  if (fs.exists(PREFERENCES_SERVO_FILE)) {
    Serial.printf("Updating %s\r\n", PREFERENCES_SERVO_FILE);
  } else {
    Serial.printf("Creating %s\r\n", PREFERENCES_SERVO_FILE);
  }
  File file = fs.open(PREFERENCES_SERVO_FILE, FILE_WRITE);
  static char json_response[1024];
//  sensor_t * s = esp_camera_sensor_get();
  char * p = json_response;
  *p++ = '{';
  p+=sprintf(p, "\"s_horizon_min\":%i,", s_horizon_min);
  p+=sprintf(p, "\"s_horizon_max\":%i,", s_horizon_max);
  p+=sprintf(p, "\"s_horizon_cen\":%i,", s_horizon_cen);
  p+=sprintf(p, "\"s_vertic_min\":%i,", s_vertic_min);
  p+=sprintf(p, "\"s_vertic_max\":%i,", s_vertic_max);
  p+=sprintf(p, "\"s_vertic_cen\":%i ", s_vertic_cen);
  *p++ = '}';
  *p++ = 0;
  file.print(json_response);
  file.close();
  dumpPrefsServo(SPIFFS);
}

/*
 * procedura kontroli płynnego ruchu serw
 */
void serwaMiekkiRuch(void)
{
#define SERVA_MR_INTERWAL_US  50*1000  
  
  if (serva_miekki_ruch_cnl==false) return;//serwa stop 

  if (servo1_ruch==true)
  {
    if (serwaMR_krok_ruchu(&servo1Pos)==true) servo1_ruch =false;
  }
  if (servo2_ruch==true)
  {
    if (serwaMR_krok_ruchu(&servo2Pos)==true) servo2_ruch =false;
  }
  if ((servo1_ruch==false)&&(servo2_ruch==false)) serva_miekki_ruch_cnl =false; 
  else TimerLib.setTimeout_us(serwaMiekkiRuch, SERVA_MR_INTERWAL_US);
}
/*
 * płynny ruch serw krok ruchu
 * p_servoPos -docelowy kąt wychylenia serwa
 * wy: true -kąt osiągnięty
 */
 bool serwaMR_krok_ruchu(int *p_servoPos)
 {
#define SERVA_MR_KROK_US  5
  
char servo_num;
int servoPosAng_akt,servoMicrosec_akt;
static signed char kierunek_korekty=0;
  
  if (p_servoPos ==&servo1Pos) servo_num=1;
  else servo_num=2;

  if (servo_num==1)
  {
    servoPosAng_akt =servo1.read();
    servoMicrosec_akt =servo1.readMicroseconds();
  }
  else
  {
    servoPosAng_akt =servo2.read();
    servoMicrosec_akt =servo2.readMicroseconds();
  }

  if (servoPosAng_akt==*p_servoPos) 
  {
    kierunek_korekty=0;
    return true;//kąt osiągnięty
  }
  if (servoPosAng_akt< *p_servoPos) 
  {
    if (kierunek_korekty<0)
    {
      kierunek_korekty=0;
      return true;//kąt osiągnięty
    }
    servoMicrosec_akt+= SERVA_MR_KROK_US;
    kierunek_korekty =1;
  }
  else 
  {
    if (kierunek_korekty>0)
    {
      kierunek_korekty=0;
      return true;//kąt osiągnięty      
    }
    servoMicrosec_akt-= SERVA_MR_KROK_US;
    kierunek_korekty =-1;
  }
  if (servo_num==1) servo1.writeMicroseconds(servoMicrosec_akt);
  else servo2.writeMicroseconds(servoMicrosec_akt);

//  if (digitalRead(servaOnOff1)==0) digitalWrite(servaOnOff1, HIGH);
//  else digitalWrite(servaOnOff1, LOW);

  return false;
 }
/*
 * plynny ruch serw start
 * p_servoPos -docelowy kąt wychylenia serwa
 * 
 */
 void servaMiekkiRuch_start(int *p_servoPos)
 {
  serva_mr_interwal =millis();
  
  if (p_servoPos ==&servo1Pos)  servo1_ruch =true;
  else  servo2_ruch =true;
  serva_miekki_ruch_cnl =true;
  TimerLib.setTimeout_us(serwaMiekkiRuch, SERVA_MR_INTERWAL_US);
 }
